/* Copyright (C) 2011-2013 Felix Salfelder, Lars Hedrich
 * Author: Felix Salfelder
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *------------------------------------------------------------------
 * a remote control socket. used by verification tools
 */
#define ADD_VERSION // signon?
#include "globals.h"
#include "u_status.h"
#include <unistd.h>
#include "u_prblst.h"
#include "u_cardst.h"
#include "u_nodemap.h"
#include "e_elemnt.h"
#include "e_storag.h"
#include "e_subckt.h"
#include "d_subckt.h"
#include "u_sock.h"
#include "s_ddc.h"
#include "io_error.h"
#include "resolv.h"
#include "s__.h"
#include "io_matrix.h"
#include "m_matrix_extra.h"
#include <iomanip>
#include "d_cap.h"
#include <fcntl.h>
/*--------------------------------------------------------------------------*/
#define userinfo( a,b,c,d,e,f )
/*--------------------------------------------------------------------------*/
using namespace SOME_CAP_HACK; // FIXME. (maybe use STORAGE device interface?)
/*--------------------------------------------------------------------------*/
namespace {
/*--------------------------------------------------------------------------*/
typedef union {
  double double_val;
  int32_t   int_val;
} di_union_t;

/*--------------------------------------------------------------------------*/
class SOCK : public DDC_BASE { //
public:
  explicit SOCK();
  ~SOCK();
  string status()const;
protected:
  void	options(CS&, int x=0);
private:
  void	sweep();
  void	sweep_recursive(int);
  void	first(int);
  bool	next(int);
  void send_matrix();
  void undo_time_step();
  void tr_reject();
  static unsigned tr_steps_rejected_;
protected:
  enum {DCNEST = 4};
  int _n_sweeps;
//  typedef void (*p)(double);
  //ELEMENT* (_pushel[DCNEST]);	/* pointer to thing to sweep, dc command */
  CARDSTASH _stash[DCNEST];	/* store std values of elements being swept */
  bool _loop[DCNEST];		/* flag: do it again backwards */
  bool _reverse_in[DCNEST];	/* flag: sweep backwards, input */
  bool _reverse[DCNEST];	/* flag: sweep backwards, working */
  bool _cont;			/* flag: continue from previous run */
  TRACE _trace;			/* enum: show extended diagnostics */
  enum {ONE_PT, LIN_STEP, LIN_PTS, TIMES, OCTAVE, DECADE} _stepmode[DCNEST];
  static double temp_c_in;	/* ambient temperature, input and sweep variable */
  bool _dump_matrix;
  unsigned port;
  double* U;
  double* CU;
  double* CUTCU;
public:
  void	do_it(CS&, CARD_LIST*);
private:
  void	setup(CS&);
  void fillnames( const CARD_LIST* scope);
  void findcaps( CARD_LIST* scope);
  void cap_prepare();
  void cap_reset();
  std::vector<IString> var_namen_arr;
  std::vector<COMPONENT*> _caplist; // FIXME: use cardlist
  std::vector<double> _cap_ic_list; // for storing the ic values from the
                                    // netlist 
  CARDSTASH* _capstash;
  uint16_t var_namen_total_size;

private: //vera stuff.
  void main_loop();
  void verainit(unsigned, unsigned, unsigned);
  void verakons();
  bool relax_kons(bool converged_in);
  bool has_sol_nan(string message);

  void veraop();
  void verainit_reply();
  void verakons_reply();
  void veraop_reply();
  void ev_iter();
  void set_param();
  unsigned transtep(unsigned init, double dt);
  void transtep_reply(unsigned, bool eol=true);
  void transtep_gc_reply(unsigned);

  Ichar* var_names_buf;

  unsigned _verbose;
  size_t total;
  size_t n_inputs() const{return _input_names.size();}
  std::vector<IString> _input_names;
  std::vector<ELEMENT*> _input_devs;
  unsigned n_vars;
  unsigned n_vars_square;
  uint16_t n_eingaenge;
  uint16_t length;

  SocketStream stream;
  //unsigned BUFSIZE;
  unsigned n_bytes;

  uint16_t _error; // transport between method and method_reply
  double _dthack; // same

  double *dc_werteA,*dc_loesungA,*kons_loesungA,*kons_residuumA;

  int channel;
  int frame_number;
  Socket* socket;
  short unsigned _port_range;
  string _port;   // kommt ueber die Uebergabeparamter port
                  // globale Variable daher (default: port=1400)
  bool _veraopuic; // use uic=true in veraop if given at command line 
  bool _client_mode;
  bool _server_mode;
  unsigned _bufsize;
  bool _bigarg;
  bool _binout;
  string _host;
  int reuseaddr;
  struct sockaddr_in sin;

  double *matrixg, *matrixc,*vectorq;

  static const int printlevel=0;
};
/*--------------------------------------------------------------------------*/
unsigned SOCK::tr_steps_rejected_ = 0;
double	SOCK::temp_c_in = 0.;
/*--------------------------------------------------------------------------*/
void SOCK::do_it(CS& Cmd, CARD_LIST* Scope)
{
  trace0("SOCK::do_it");
  _scope = Scope;
  _sim->_time0 = 0.;
  //_sim->set_command_ddc();
  _sim->set_command_dc();
  _sim->_phase = p_INIT_DC;
  //_sim->_ddc = true;
  ::status.ddc.reset().start();
  _sim->_temp_c = temp_c_in;
  _dump_matrix = 0;
  reuseaddr = 0;
  _port = "1400";
  _veraopuic=false;
  _port_range = 1;
  _client_mode = false;
  _host = "localhost";

  command_base(Cmd);

  //cleanup
  cap_reset();
}
/*--------------------------------------------------------------------------*/
SOCK::SOCK() :DDC_BASE(),
   _n_sweeps(1),
   _cont(false),
   _trace(tNONE)
{

  for (int ii = 0; ii < DCNEST; ++ii) {
    _loop[ii] = false;
    _reverse_in[ii] = false;
    _reverse[ii] = false;
    _step[ii] = 0.;
    _linswp[ii] = true;
    _zap[ii]=NULL;
    _stepmode[ii] = ONE_PT;
  }
  temp_c_in=OPT::temp_c;
  _out=IO::mstdout;
}
/*--------------------------------------------------------------------------*/
void SOCK::setup(CS& Cmd)
{
  _sim->_uic = false;
  trace0("SOCK::setup");
  _cont = false;
  _trace = tNONE;
  _out = IO::mstdout;
  _out.reset(); //BUG// don't know why this is needed */
  bool ploton = IO::plotset  &&  plotlist().size() > 0;

  options(Cmd);

  if (Cmd.more()) {
  }else{
  }
  Cmd.check(bWARNING, "what's this?");

  _sim->_uic = true;
  CKT_BASE::_sim->init();

  IO::plotout = (ploton) ? IO::mstdout : OMSTREAM();
  initio(_out);

  _error = 0; /* verainit(v_flag, n_inputs, &n_vars, charbuf, &length); */
  n_vars = static_cast<uint16_t>( _sim->_total_nodes) ; // _sim->total_nodes doesnt include gnd
  // do this also in sweep to initialize the socket streaM
  var_namen_arr.resize( n_vars, string("unset"));
//  var_namen_arr[0]="0";
  fillnames( &CARD_LIST::card_list );
  n_vars_square = n_vars * n_vars;

  findcaps(&CARD_LIST::card_list);

  // BUG! this kills cap bm's
  cap_prepare(); // attach value common if !has_common. stash old common

#ifndef NDEBUG
    for (unsigned i=0; i < n_vars; i++)
      trace0("name: " + var_namen_arr[i]);
#endif

  assert(_n_sweeps > 0);
  _sim->_freq = 0;

  // not implemented. need to queue sources properly (CARDLIST::q_hack..?)
  if(OPT::prequeue) { incomplete();
    error(bDANGER, "prequeueing is incomplete, disabling\n");
    OPT::prequeue = false;
  }else{
  }

  IO::error.detach(stdout);
  IO::error.attach(stderr);
}
/*--------------------------------------------------------------------------*/
void SOCK::options(CS& Cmd, int Nest)
{
  trace0("SOCK::options... ");

  _sim->_uic = _loop[Nest] = _reverse_in[Nest] = false;
  _port = "1400";
  _veraopuic=false;
  _bufsize = BUFSIZE;
  _bigarg = true;
  _binout = true;
  unsigned here = Cmd.cursor();
  do{
    ONE_OF
      || (Cmd.match1("'\"({")	&& ((Cmd >> _step_in[Nest]), (_stepmode[Nest] = LIN_STEP)))
      || (Cmd.is_float()	&& ((Cmd >> _step_in[Nest]), (_stepmode[Nest] = LIN_STEP)))
      || (Get(Cmd, "*",		  &_step_in[Nest]) && (_stepmode[Nest] = TIMES))
      || (Get(Cmd, "+",		  &_step_in[Nest]) && (_stepmode[Nest] = LIN_STEP))
      || Get(Cmd, "b{ufsize}",    &_bufsize)
      || Get(Cmd, "c{ontinue}",   &_cont)
      || Get(Cmd, "port" ,        &_port)
      || Get(Cmd, "listen{port}", &_port)
      || Get(Cmd, "bigarg",       &_bigarg)
      || Get(Cmd, "binout",       &_binout)
      || Get(Cmd, "host" ,        &_host)
      || Get(Cmd, "uic" ,        &_veraopuic)
      || Get(Cmd, "tr{s}",        &_do_tran_step)
      || Get(Cmd, "dm",           &_dump_matrix)
      || Get(Cmd, "client",       &_client_mode)
      || Get(Cmd, "server",       &_server_mode)
      || Get(Cmd, "dt{emp}",	  &temp_c_in,   mOFFSET, OPT::temp_c)
      || Get(Cmd, "lo{op}", 	  &_loop[Nest])
      || Get(Cmd, "re{verse}",	  &_reverse_in[Nest])
      || Get(Cmd, "te{mperature}",&temp_c_in)
      || (Cmd.umatch("tr{ace} {=}") &&
	  (ONE_OF
	   || Set(Cmd, "n{one}",      &_trace, tNONE)
	   || Set(Cmd, "o{ff}",       &_trace, tNONE)
	   || Set(Cmd, "w{arnings}",  &_trace, tUNDER)
	   || Set(Cmd, "i{terations}",&_trace, tITERATION)
	   || Set(Cmd, "v{erbose}",   &_trace, tVERBOSE)
	   || Cmd.warn(bWARNING,
		       "need none, off, warnings, iterations, verbose")
	   )
	  )
      || _out.outset(Cmd);
  }while (Cmd.more() && !Cmd.stuck(&here));
  ::error(bWARNING, "SOCK::options uic for veraop set to %s\n",_veraopuic?"True":"False");
  
}
/*--------------------------------------------------------------------------*/
static void register_status();
/*--------------------------------------------------------------------------*/
void SOCK::sweep()
{
  register_status();
  // later... FIXME
  //
  frame_number = 0;
  n_bytes = 0;
  trace1("SOCK::do_it", _port);

  // Calculate buffersize
  n_vars = _sim->_total_nodes;
  // _sim->total_nodes doesnt include gnd
  n_vars_square = n_vars * n_vars;
  _bufsize = n_vars_square * 3 * (unsigned)sizeof(double) + 32;
  // 3 quadratic matrices * doubles + some safety margin (BUG)

  if (_client_mode){
    trace1("bufsize Client ", _bufsize);
    socket = new ClientSocket(Socket::TCP, _port, _host, _bufsize);
    trace1("connected to "+ _host, _port );
    stream = *socket;
  } else if (_server_mode) {
    stringstream p(_port);
    uint16_t _port_;
    p >> _port_;
    trace1("bufsize Server ", _bufsize);
    socket = new ServerSocket(Socket::TCP, _port_, 1u, _bufsize);
    ServerSocket* sock=prechecked_cast<ServerSocket*>(socket);

    trace0("SOCK::do_it waiting");
    stream = sock->listen();
  }else{
    fflush( stdout );
    fflush( stdin );
    trace0("SOCK::sweep simple i/o");
    socket = NULL;
    trace1("bufsize Stdin ", _bufsize);
    if (!_binout){
      int devnull=open("/dev/null", O_WRONLY);
      trace2("Socket stream for dev null" ,devnull, STDIN_FILENO);
      stream = SocketStream( devnull, STDIN_FILENO, _bufsize);
    }else{
      stream = SocketStream( STDOUT_FILENO, STDIN_FILENO, _bufsize);
    }
    stream << "gnucap sock ready";
  }

  main_loop();

  delete socket;

  _out << "\n"; // (good idea?)

  IO::error.detach(stderr);
  IO::error.attach(stdout); // BUG. does not seem to work?!
  return;
}
/*--------------------------------------------------------------------------*/
string SOCK::status()const
{
  // incomplete();
  return "sock: tr rejected=" + ::to_string(tr_steps_rejected_) + "\n";
}
/*--------------------------------------------------------------------------*/
// void SOCK::sweep_recursive(int )
// {
//   assert(false);
// }
/*--------------------------------------------------------------------------*/
// fetch names from circuit recursively. fill into local vector.
void SOCK::fillnames( const CARD_LIST* scope){
  trace0("SOCK::fillnames");

  const NODE_MAP * nm = scope->nodes();
  for (NODE_MAP::const_iterator i = nm->begin(); i != nm->end(); ++i) {
    if (i->first != "0") {
      if (const NODE* a= dynamic_cast<const NODE*>(i->second)){
        stringstream s;
        IString myname(a->long_label());
        var_namen_arr[a->matrix_number()-1] = myname;
        var_namen_total_size = static_cast<uint16_t>( var_namen_total_size + static_cast<uint16_t>(myname.length()) + 1 );
      }
    }else{
    }
  }

  for (CARD_LIST::const_iterator i = scope->begin(); i != scope->end(); ++i) {
    if(const COMPONENT* s = dynamic_cast<const COMPONENT*>(*i)){
      if (!s->is_device()){
      }else if ( s->subckt() ) {
        fillnames( s->subckt() );
      }
    }else{
    }
  }
}
/*--------------------------------------------------------------------------*/
void SOCK::findcaps( CARD_LIST* scope){
  int nrealcaps=0;
  for (CARD_LIST::iterator i = scope->begin(); i != scope->end(); ++i) {
    if ( COMPONENT* c = dynamic_cast< COMPONENT*>(*i) ) {
      if (c->is_device() && c->has_memory()){
        // c->long_label().find("C1")==0 ||
       // if (1==1 || c->long_label().find("CGBDum")==std::string::npos
       //     && c->long_label().find("CSBDum")==std::string::npos
       //     && c->long_label().find("CDBDum")==std::string::npos) {
       // if (c->long_label().find("CGBDum")==std::string::npos &&
       //     c->long_label().find("CSBDum")==std::string::npos) {
        if ( 1==1) {
        //ELEMENT* d = prechecked_cast<ELEMENT*>(c);

        trace2("found cap", c->long_label(),c->dev_type());
        ::error(bDEBUG+_verbose,"SOCK::findcaps: found cap %s \n", c->long_label().c_str());
        _caplist.push_back( c );
        if (STORAGE* d = dynamic_cast<STORAGE*>(c)) { // only built in c's
          nrealcaps++;

          // not for behavioral models.
          // if we want to have them, than we have to uic for behavioral models
          // we have to implement set_ic(), set__ic() in e_adms.h in
          // gnucap-adms
          trace2("initial cond",d->has_ic(),d->set__ic());
          if (d->has_ic()) {
            double val=*(d->set__ic());
            _cap_ic_list.push_back(val);
            trace2("cap has ic=",c->long_label(),val);
            ::error(bDEBUG+_verbose,"cap has ic %s %g\n", c->long_label().c_str(),
                    val);
          } else {
            _cap_ic_list.push_back(NOT_INPUT);
            ::error(bDEBUG+_verbose,"cap has no ic %s \n", c->long_label().c_str());
          }
        } else {
          _cap_ic_list.push_back(NOT_INPUT);
        }    
       } else {
          trace1("ignoring cap",c->long_label());
       }
      }
    }
    if (!(*i)->is_device()){
    } else if ( BASE_SUBCKT* s = dynamic_cast< BASE_SUBCKT*>(*i) ) {
      trace1("going down", s->long_label());
      findcaps( s->subckt() );
    }
  }
  // if (nrealcaps==0) {
  //   ::error(bWARNING,"SOCK::findcaps: No caps to be set "
  //           "by uic found, no state variables useable\n");
  // } else {
  //   ::error(bDEBUG,"SOCK::findcaps: Number of "
  //           "caps found: %d \n",nrealcaps );
  // }

}
/*--------------------------------------------------------------------------*/
static SOCK p2;
static DISPATCHER<CMD>::INSTALL d2(&command_dispatcher, "sock", &p2);
/*--------------------------------------------------------------------------*/
DISPATCHER<CKT_BASE>::INSTALL* _status=NULL;
static void register_status()
{
  if(!_status){
    _status = new DISPATCHER<CKT_BASE>::INSTALL(&status_dispatcher, "sock", &p2);
  }else{ untested();
  }
}
/*--------------------------------------------------------------------------*/
}
/*--------------------------------------------------------------------------*/
static unsigned argc(unsigned opcode)
{
  switch(opcode){
    case 51:
      return 3;
    case 52: itested();
      return 0;
    case 53:
      return 0;
    case 104: itested();
    case 102:
      return 1;
    default:
      return 0;
  }
}
/*--------------------------------------------------------------------------*/
void SOCK::main_loop()
{
  trace0("SOCK::main_loop");

  uint16_t opcode=-1;
  uint16_t arg[3];
  unsigned char tmp;
  arg[0] = -1;
  bool init_done=false;

  while(true) {
    trace0("SOCK::main_loop waiting for opcode");

    if(_bigarg){
      stream >> tmp >> 7; // sic!
      opcode = tmp;
      for(unsigned i=0; i<argc(opcode); ++i){
        stream >> arg[i] >> 6;
      }
    }else{
      stream >> opcode;
      stream >> arg[0];
      stream >> arg[1];
      stream >> arg[2];
    }

    _sim->_mode = s_SOCK; // nonsense.
                          // use respective built-in mode!
                          //
    if(_bigarg) {
      ::error(bDEBUG, "sock opcode %d\n", opcode);
    }else{
      ::error(bDEBUG, "sock opcode %d %d %d %d\n", opcode, arg[0], arg[1], arg[2]);
    }

    double dt;
    unsigned status;
    switch (opcode) {
      case '\0':
        return;
      case '3': // 51
        if(init_done) throw Exception("init twice??");
        verainit(arg[0], arg[1], arg[2]);
        verainit_reply();
        init_done=true;
        break;
      case '4':  // 52
        veraop();
        veraop_reply();
        break;
      case '5':  // 53
        verakons();
        verakons_reply();
        break;
      case 101: untested();
        ev_iter();
        break;
      case 102:
        stream >> dt;
        status = transtep(arg[0], dt);
        transtep_reply(status);
        break;
      case 103: itested();
        set_param();
        break;
      case 104: itested();
        stream >> dt;
        status = transtep(arg[0], dt);
        transtep_gc_reply(status);
	break;
      default:
        ::error(bDANGER, "unknown opcode %i\n", opcode);
        throw Exception("unknown opcode");
        break;
    }
    _sim->reset_iteration_counter(iPRINTSTEP); // used by solve_with_homotopy
                                               // reset in outdata (not called)
  }
}
/*--------------------------------------------------------------------------*/
// very clever way to transfer strings.
static void putstring8(SocketStream* s, const IString x)
{
  const Ichar* A = x.c_str();

  while(*A!='\0'){
    *s << A->to_char();
    *s << A->to_char();
    *s << A->to_char();
    *s << A->to_char();
    *s << A->to_char();
    *s << A->to_char();
    *s << A->to_char();
    *s << A->to_char();
    A++;
  }
}
/*--------------------------------------------------------------------------*/
void SOCK::verainit(unsigned verbose, unsigned n_in, unsigned length)
{
  if (verbose > 255 ) {
    _verbose=0;
  } else {
    _verbose = verbose;
  }
  ::error(bWARNING,"SOCK::verainit: _verbose %d \n",_verbose);

  char input_namen[length+1];
  unsigned here =0;
  unsigned n=0;

  _input_names.resize(n_in);
  _input_devs.resize(n_in);
  trace3("verainit: ", verbose,n_inputs(),length);
  assert(n_inputs()==n_in);

  for (unsigned i=0; i < length; i++) {
    stream >> input_namen[i] >> 7;
    if(input_namen[i] == '\t'){
      input_namen[i] = 0;
      trace1("input_namen", string(input_namen+here));
      trace5("input_namen",input_namen[i-2],input_namen[i-1], input_namen[i-0],here,i );
      _input_names[n] = string(input_namen+here);
      here = i+1;

      trace1("looking out for", _input_names[n]);
      CARD_LIST::fat_iterator ci = findbranch(_input_names[n], &CARD_LIST::card_list);
      if (ci.is_end()){
        throw Exception("cannot find voltage source \"" + _input_names[n] +"\", giving up");
      }
      trace0("found input device");
      ELEMENT* d = prechecked_cast<ELEMENT*>(*ci);
      if (!d){
        throw Exception("not something we can use as source, \"" +  _input_names[n] +"\", giving up");
      }
      _input_devs[n] = d;
      assert(_input_devs[n]);

      unsigned ii=n;
      // see s_cd. FIXME: revert.
      _stash[ii] = _input_devs[ii];
      _input_devs[ii]->inc_probes();
      _input_devs[ii]->set_value(_input_devs[ii]->value(),0);
      _input_devs[ii]->set_constant(false); // <= insufficient for PREQEUE

      ++n;
    }
  }

  //trace0("input_namen " + string(input_namen) );
  total = (unsigned) (length+4);
  assert(stream.bufsize() >= total);

  assert(!var_names_buf);
  var_names_buf = (Ichar*) malloc( n_vars * 128 * sizeof(char));
  strcpy((char*)var_names_buf, "");
  for (unsigned i=0; i < n_vars; i++)
  {
    trace1("SOCK::verainit ", var_namen_arr[i]);
    strcat((char*)var_names_buf, (const char*)var_namen_arr[i].c_str());
    strcat((char*)var_names_buf, "\t");
  }
  //length = static_cast<uint16_t>( strlen(var_names_buf) );
  // userinfo(1,"vera_titan_ak","Variablennamen %s\n",var_names_buf);
  OPT::incmode =false; // Paranoid: never use an inc_mode in simulation
  _sim->set_inc_mode_no();

}
/*--------------------------------------------------------------------------*/
void SOCK::veraop()
{
  trace1("SOCK::veraop",n_vars);
  total = n_vars;
  _sim->set_command_dc();
  assert(stream.bufsize() >= total);

  dc_werteA = (double*) malloc(sizeof(double)*n_vars);
  trace1("fetching ",n_vars);
  assert(_sim->vdc()[0] == 0 );
  for (unsigned i=0; i < n_inputs(); i++) {
    double d;
    stream >> d;
    trace3("SOCK::veraop setting input", _input_devs[i]->long_label(), i, d);
    _input_devs[i]->set_value(d);
  }

  // now resetting the ic conditions to initial ones
  // otherwise a call to verakons would overwrite all ic=
  // given in the netlist. We need that
  // only in case we use uic at command line for VCOs and Integrators
  // 
  for( unsigned i = 0; i < _caplist.size(); i++) {
    trace1("SOCK::kons",_caplist[i]->long_label());
    //STORAGE* d = prechecked_cast<STORAGE*>(_caplist[i]);
    if (STORAGE* s = dynamic_cast<STORAGE*>(_caplist[i])) { // only built in c's
      if( ELEMENT* d = prechecked_cast<ELEMENT*>(_caplist[i])) { 
        double val=-1;
        if(_cap_ic_list[i]!=NOT_INPUT) {
          trace3("resetting ic from to netlist value cap ic=",
                 d->long_label(),_cap_ic_list[i],s->has_ic());
          if (double* valp=s->set__ic()) {
            val=*valp; }
          ::error(bWARNING+_verbose,
                  "resetting ic from cap %s from value %g to "
                  "value ic=%g has_ic %d\n",
                  d->long_label().c_str(),val, _cap_ic_list[i],
                  s->has_ic()?1:0);
          d->set_ic(_cap_ic_list[i]);
        } else {
          if (double* valp=s->set__ic()) {
            val=*valp; }
          trace3(" ic from to netlist value cap ic=",
                 d->long_label(),_cap_ic_list[i],s->has_ic());
          d->set_ic(NOT_INPUT); // this is needed if in between vera kons has
                    // been used. because that sets other values ot the ic. 
          ::error(bDEBUG+_verbose,
                  "ic from cap %s not reset. ic val =%g ic val in "
                  "remembered list : %g has_ic %d \n",
                  d->long_label().c_str(),val,
                  _cap_ic_list[i],s->has_ic()?1:0 );
          trace3(" ic from to netlist value cap ic=",
                 d->long_label(),_cap_ic_list[i],s->has_ic());
        } 
      }
    }
  }


  _error = 0; /* veraop(sweep_val, x_new, G, C); */

  // ================do_dc========
  //  _sim->_uic = false;
  _sim->_uic = _veraopuic; // in general it should be false to
                           // not user the old values stored in
                           // the caps from may be previous runs
                           // of verakons.
                           // for VCOs and Integrators it could be set
  // true by commandline uic in .sock command generated form vera by option uic
  _sim->_bypass_ok = false;
  _sim->set_inc_mode_bad();
  OPT::ITL itl = OPT::DCBIAS;
  _sim->_phase = p_INIT_DC;
  _sim->_time0 = 0; // paranoid, the procedure
                    //solve_with_homotopy is asserting this
  _sim->_dt0 = 0; // dito
  advance_time();

  trace2("SOCK::veraop homotopy", _sim->_phase, _sim->_mode);
  if(_dump_matrix) {
    _trace=tVERBOSE;
  }
  CARD_LIST::card_list.tr_begin();  // hier muesste eigentlich eine dc hin.
  if(printlist().size()) {
    head(0,0," ");
  }
  bool converged = false;
  try{
    converged = solve_with_homotopy(itl,_trace);
  }catch( Exception e) {
    ::error(bDANGER, "homotopy failed\n");
    throw e;
  }
  _sim->_uic = false; //set back 

  if(!converged){ itested();
    _error = 1;
    if(_dump_matrix) {
      _out << " not converged\n";
    }
  }else{
    if(_dump_matrix) {
      _out << " converged\n";
    }
    ::status.accept.start();
    _sim->set_limit();
    trace0("SOCK::veraop, accepting");
    CARD_LIST::card_list.tr_accept();
    ::status.accept.stop();

    _sim->keep_voltages();
  }
  //========================

  // Die Variablenwerte stehen schon durch solve_system in var_werte
  // in dc_sysA und muessen noch nach A kopiert werden:
  //
  //A->var_werte = x_neu;
  //free(dc_loesungA);
  //
  //        A->var_werte = dc_loesungA;
  // diff_werte, par_werte muessen auch gesetzt sein!
  // in diesem Falle sind sie per default = 0
  //
  //  solve  somthing
  //
  //  A->eval_lin_gl(dc_werteA,&matrixg,&matrixc,&vectorq);

  trace1("matrix", 1);
}
/*--------------------------------------------------------------------------*/
void SOCK::verakons()
{
  //        n_eingaenge == #caps?
  _sim->set_command_dc();
  trace1("SOCK::kons", _caplist.size());
  _sim->_time0 = 0; // just assert?
  _sim->_dt0 = 0; // just assert?
  advance_time();
  _sim->_phase = p_INIT_DC;
  total =  n_eingaenge + n_vars + 1;

  for (unsigned i=0; i < n_inputs(); i++) {
    double d;
    stream >> d;
    trace3("SOCK::verakons setting input", _input_devs[i]->long_label(), i, d);
    _input_devs[i]->set_value(d);
  }

  assert(_sim->_v0);
  for (unsigned i=1; i <= n_vars; i++) {
    stream >> _sim->_v0[i];
    //    trace2("SOCK::kons start ", i,  _sim->_v0[i] );
  }
  if (printlist().size()) {
    outdata(0., ofPRINT);
  }
  _sim->keep_voltages(); // v0->vdc

  _error = 0; /* verakons(Dwork, x_new, q_dot, G, C); */
  //	n_vars = A->n_var;

  for( unsigned i = 0; i < _caplist.size(); i++) {
    //trace1("SOCK::kons",_caplist[i]->long_label());
    ::error(bDEBUG+_verbose,
            "SOCK::kons keep ic for %s \n",_caplist[i]->long_label().c_str());
    _caplist[i]->keep_ic(); // latch voltage applied to _v0
    // this calls set_ic() and sets has_ic() to a value, which means it has
    // an initial condition. Which was a problem. Now fixed. in op() it will
    // be reset by writing NOT_INPUT into the ic. 
    _caplist[i]->set_constant(true); // maybe not a good idea at all.
    _caplist[i]->q_eval();		// so it will be updated
  }
  //
  //================do_dc========
  _sim->_uic = true;
  _sim->_bypass_ok = false;
  _sim->set_inc_mode_bad();

  OPT::ITL itl = OPT::DCBIAS;

  if(printlist().size()) {
    head(0,0," ");
  }
  CARD_LIST::card_list.tr_begin();
  bool converged = false;
  try{
    for( unsigned i = 0; i < _caplist.size(); i++) {
      //trace1("SOCK::kons",_caplist[i]->long_label());
      if(_verbose > 10) {
        ::error(bDEBUG+_verbose,"SOCK::kons: setting cap %s \n",
                (_caplist[i]->long_label()).c_str());
      }
      // assert(!_caplist[i]->is_constant()); // NO caps are const, when they are in VS mode.
      _caplist[i]->do_tr(); // so it will be updated. is this sufficient?
			     // is it sufficient to only queue caps?
    }
    trace1("SOCK::verakons: solving",_trace);
    converged = solve(itl,_trace);
    //    solve(OPT::TRHIGH,_trace);
    //    solve_with_homotopy(itl,_trace);
    // homotopy is to much effort calling
    // procedures must catch problems with convergence.
  }catch(Exception e) { untested();
    ::error(bDANGER+_verbose, "solve failed in verakons\n");
    throw e;
  }
  if(!converged){
    ::error(bWARNING, "s_sock::verakons: solve did not converge\n");
    converged = solve_with_homotopy(itl,_trace);
  }
  if(!converged){
    _error = 1;
  }
  ::status.accept.start();
  assert(_sim->_uic);
  CARD_LIST::card_list.tr_accept();
  ::status.accept.stop();
  
  if(_dump_matrix){
    _out << "AA u,dq=i: \n";
    for (unsigned i=0; i <= n_vars; i++)         /* Variablen-Werte */
    {
      _out << _sim->_v0[i] << "," << _sim->_i[i] <<  "\n" ;
    }
  }

  if ( 1==0) {
    // ======  a mini step for relaxing the strong voltages =====
    converged=relax_kons(converged);
  } 

  if (printlist().size()) {
    outdata(_sim->_time0, ofPRINT);
  }
  _sim->_mode = s_SOCK; // for now.

  _sim->keep_voltages();
  _sim->zero_currents();
  _sim->set_inc_mode_no();
  if(_dump_matrix){
    _out << "B1 u,dq=i: \n";
    for (unsigned i=0; i <= n_vars; i++)         /* Variablen-Werte */
    {
      _out << _sim->_v0[i] << "," << _sim->_i[i] <<  "\n" ;
    }
  }

  // vera wants just cap stamps
  if(converged) {
    //   final step of solve() will call evaluate models() which
    //   calculates with do_tr the values of the caps and
    //   the tr_load() here will stamp the values in the matrix.
	if (_sim->_loadq.empty()) {
	  trace0("loadq  empty: clearing loadq");
	  set_flags(); // copied from s__solve.cc because if the
	  // solving succeeded in last iteration, this won't be done 
	  clear_arrays();
	  finish_building_evalq();
	  
	  _sim->count_iterations(iPRINTSTEP);
	  _sim->count_iterations(iSTEP);
	  _sim->count_iterations(_sim->_mode);
	  _sim->count_iterations(iTOTAL);
	  
	  evaluate_models();
	}
    if(_dump_matrix){
      _out << "B2 u,dq=i: \n";
      for (unsigned i=0; i <= n_vars; i++)         /* Variablen-Werte */
      {
        _out << _sim->_v0[i] << "," << _sim->_i[i] <<  "\n" ;
      }
    }
    //load_matrix();
    for (unsigned i = 0; i < _caplist.size(); i++) { itested();
      CARD* c = prechecked_cast<CARD*>(_caplist[i]);
      assert(c);
      _sim->_damp = 1.; // need raw stamps
      //      if (c->long_label().find("C1")==0) {
      //if (c->long_label().find("CDBDum")==std::string::npos ){
      if(n_vars > 20) {
        double delta_i = -1.* _sim->_i[20];
        c->tr_load(); // this can be done only when converged
        delta_i+= _sim->_i[20];
        trace3("verakons, loaded cap _i [20]", c->long_label(),delta_i,_sim->_i[20]);
      } else {
        c->tr_load(); // this can be done only when converged
        trace1("verakons, loaded cap ", c->long_label());
      }
      
      //}
    }
    if(_dump_matrix){
      _out << "B3 u,dq=i: \n";
      for (unsigned i=0; i <= n_vars; i++)         /* Variablen-Werte */
      {
        _out << _sim->_v0[i] << "," << _sim->_i[i] <<  "\n" ;
      }
  }

  }else{itested();
    ::error(bWARNING, "s_sock::verakons: solve did not converge even with homotopy\n");
    // avoid double loads. in case of _error
	  // If solve or solve with homotopy did not converge,
	  // it will not increment the iTOTAL counter and
	  // we would get a double tr_load_source if this is executed.
    _error=1;
  }
  if(_dump_matrix){
    _out << "C u,dq=i: \n";
    for (unsigned i=0; i <= n_vars; i++)         /* Variablen-Werte */
    {
      _out << _sim->_v0[i] << "," << _sim->_i[i] <<  "\n" ;
    }
  }
	
  for( unsigned i = 0; i < _caplist.size(); i++) {
	/// oops. maybe the next command is tran?!
	// (this is a hack!!)
	_caplist[i]->set_constant(false);
  }
	
}
/*--------------------------------------------------------------------------*/
#undef rescale
#define normalize
void SOCK::ev_iter()
{
  trace1("SOCK::ev_iter", n_vars);
  static double* aug;
  double lambda;
  if(!aug){ untested(); // quick hack!
    assert(_sim->_aa.size()==n_vars);
    aug = new double[n_vars*2+1];
  }
#ifdef DO_TRACE
  for (unsigned i=0; i < 1+2*n_vars; i++){ aug[i]=i; }
#endif
  _sim->_aa = -(_sim->_acx.real());
  assert(_sim->_aa.size()==n_vars);
  double tmp[n_vars+1];
  double xiin[n_vars+1];
  double norm = 0;
  for (unsigned i=0; i < n_vars; i++) {
    stream >> tmp[i];
    trace1("SOCK::ev_iter got", tmp[i]);
    norm+=tmp[i]*tmp[i];
  }
#ifdef normalize
  norm=1./sqrt(norm);
#endif
  for (unsigned i=0; i < n_vars; i++) {
#ifdef normalize
    tmp[i]*=norm;
#endif
    aug[n_vars*2-i] = tmp[i];
    xiin[i]=tmp[i];
  }
#ifdef normalize
  aug[n_vars] = 0;
#else
  aug[n_vars] = norm-1;
#endif
  trace1("SOCK filled xi aa", _sim->_aa);
  untested();
  const BSMATRIX<double> C = _sim->_acx.imag();
  C.rmul(aug-1,xiin-1);
  stream >> lambda;

#ifdef rescale
  for (unsigned i=0; i < n_vars; i++) {aug[i]*=lambda;}
#endif

  _sim->_aa+= lambda * _sim->_acx.imag();

  trace0("SOCK::ev_iter rmul");
  _sim->_aa.rmul(tmp-1,xiin-1);
  untested();
  tmp[n_vars] = 0;
  for (unsigned i=0; i <= n_vars; i++){ untested();
    trace2("rhs",i,tmp[i]);
  }

  _sim->_aa.augment(aug);
  assert(_sim->_aa.size()==n_vars+1);
  trace1("SOCK augmented aa", _sim->_aa);
  _sim->_aa.lu_decomp();
  trace1("SOCK decomposed aa", _sim->_aa);
  untested();

  _sim->_aa.fbsub(tmp-1);
  for (unsigned i=0; i <= n_vars; i++){
    trace2("res",i,tmp[i]);
  }
  untested();
  trace0("SOCK::ev_iter done");
  _sim->_aa.deaugment();
  assert(_sim->_aa.size()==n_vars);
  stream << ((uint16_t)_error); stream.pad(6);
  double resnormsq = 0;
  for (unsigned i=0; i < n_vars; i++) { untested();
    resnormsq+=tmp[i]*tmp[i];
    stream << xiin[i]-tmp[i];
  }
#ifdef rescale
  resnormsq+=lambda*tmp[n_vars]*lambda*tmp[n_vars];
  stream << lambda*(1-tmp[n_vars]);
#else
  resnormsq+=tmp[n_vars]*tmp[n_vars];
  stream << lambda-tmp[n_vars];
#endif
  stream << resnormsq;
  stream << SocketStream::eol;
  trace0("SOCK::ev_iter sent");
}
/*--------------------------------------------------------------------------*/
void SOCK::set_param()
{ itested();
  string tmp[2];
  stream >> tmp[0];
  stream >> tmp[1];
  PARAM_LIST* pl = CARD_LIST::card_list.params();
  string paramcommand = tmp[0]+"={"+tmp[1]+"}";
  trace3("set_param", tmp[0], tmp[1], paramcommand);
  CS cmd(CS::_STRING, paramcommand);
  pl->parse(cmd);
  CARD_LIST::card_list.precalc_first(); // BUG? params don't propagate into sckts otherwise
  CARD_LIST::card_list.precalc_last();
}
/*--------------------------------------------------------------------------*/
enum{
  sOK=0,          // ok, returned time step indicates next prefered
  sNOCONV=1,      // no convergence for second or 3. newton iteration: hope
                  // for success with shorter time step
  sTRUNC=2,       // Truncation error (time step not accepted (to long!))
                  // restart with shorter returned time step
  sFIRSTNOCONV=3, // no convergence for first newton iteration after cons
                  // time step has no meaning, but reduction of time step should
                  // help
  sTROUBLE=99     // exception in solve.
};
/*--------------------------------------------------------------------------*/
// init=0 continue with ruse of all old data (do 1 step)
//      1 restore op and do ~5 steps (could be used directly after a not accepted
//        time step)
//      2 do cons_op followed by do ~5 step (depending on Euler,Gear,
//        Trapezoid, options)
//
// return:
// ret (see above)
// dt. positive timestep, use instead in next call.
//
//
unsigned SOCK::transtep(unsigned init, double dt)
{
  unsigned ret = sOK;
  _error = 0;
  trace3("SOCK::transtep", n_vars, init, dt);
  if (dt > 1E10 || dt < 0.) {
    ::error(bDANGER+_verbose, "SOCK::transtep: wanted time step dt %f seems to be erronouts",dt);
  }
  _sim->set_command_tran();
  _sim->restore_voltages(); //     _vt1[ii] = _v0[ii] = vdc[ii];
  _sim->_uic = false;
  _sim->_phase = p_RESTORE;

  unsigned stepno=-1;
  double reftime = _sim->_time0;

  if (dt==0.) {
    dt = OPT::dtmin;
  }else{
  }

  assert(dt);

  if (init==0) { itested();         // continue
    stepno = 1;
    ::error(bDEBUG, "transient step, at %f, dt %f\n", _sim->_time0, dt);
    _sim->_dt0 = dt;
    assert(_sim->_time0);
    _sim->_time0 += _sim->_dt0;
    // CARD_LIST::card_list.tr_restore();
  } else if (init==1) { itested(); // restore
    assert(_sim->_time0 > 0);
    //    CARD_LIST::card_list.tr_restore();
    stepno = 1;
    _sim->_dt0 = dt/stepno;
    ::error(bDEBUG, "restore step, dt %e\n", dt);
    _sim->_time0 += _sim->_dt0;
  } else if (init==2) {
    _sim->_time0 = reftime = 0;
    verakons(); // do we need it ?
    if (_error==1) {
      ret = sTROUBLE;
      ::error(bWARNING, "SOCK::transtep: kons operating point not found aborting ");
      return ret; 
    }      
    stepno = 5;
    _sim->_dt0 = dt/stepno;
    ::error(bDEBUG, "initial step, dt %e\n", dt);
    _sim->_time0 += _sim->_dt0;
  } else { unreachable();
  }


  trace2("SOCK::transtep ", _sim->_phase, _sim->_mode);

  advance_time(); // sink _time0 into devices
  _sim->set_command_tran();

  if (printlist().size()) {
    outdata(reftime, ofPRINT);
  }

  _sim->_phase = p_TRAN;

  //_sim->_bypass_ok = false;
  assert(_sim->_dt0 == dt/stepno);
  //_sim->_genout = gen();

  assert(_sim->analysis_is_tran());

  // this is too expensive here.
  //  CARD_LIST::card_list.do_tr();
  if (has_sol_nan("transtep before solve ")) return sTROUBLE;
  bool tr_converged;
  for (unsigned i = stepno; i>0; --i) {
    tr_converged = false;
    try {
      for( unsigned i = 0; i < _caplist.size(); i++) { itested();
        //trace1("SOCK::kons",_caplist[i]->long_label());
        assert(!_caplist[i]->is_constant());
        _caplist[i]->q_eval(); // so it will be updated. is this sufficient?
        // is it sufficient to only queue caps?
      }
      if (has_sol_nan("transtep before solve i:"+i)) return sTROUBLE;
      tr_converged = solve(OPT::TRHIGH, _trace);
      if (has_sol_nan("transtep after solve  i:"+i)) return sTROUBLE;
    }catch (Exception e) { incomplete();
      ret = sTROUBLE;
    }
    if (!tr_converged) { itested();
      ::error(bDEBUG, "s_sock::transtep did not converge\n");
      if (i==stepno) { itested();
        ret = sFIRSTNOCONV;
        break;
      } else { itested();
        ret = sNOCONV;
	break;
      }
    }else{
    }

    if(stepno>1) {
      /// uuh ooh HACK.
      assert(ret!=sFIRSTNOCONV);
      ::status.accept.start(); // accounting
      _sim->set_limit();
      CARD_LIST::card_list.tr_accept(); // accept solution in all components
      ::status.accept.stop();
    }
    if (printlist().size()) {
      outdata(_sim->_time0, ofPRINT);
    }
    if(i>1){
      _sim->_time0 += _sim->_dt0;
    }
  }

  TIME_PAIR time_by = CARD_LIST::card_list.tr_review();
  double time_by_error_estimate = time_by._error_estimate;
  assert(time_by_error_estimate>=0);

  // if(!tr_converged?) { ... } else
  if (time_by_error_estimate > _sim->_time0) {
    dt = time_by_error_estimate - _sim->_time0;
    ::status.accept.start();
    _sim->set_limit();
    CARD_LIST::card_list.tr_accept();
    ::status.accept.stop();
    _sim->keep_voltages(); //  vdc  = v0
    assert(dt>0);
  }else{ itested();
    double t1 = _sim->_time0 - _sim->_dt0;
    dt = time_by_error_estimate - t1;
    ::error(bTRACE,"sock step not accepted at %f, need %f\n", reftime, dt);
    //::error(bTRACE,"sock step not accepted at %g, need %g\n", reftime, dt);
    //assert(dt);
    ret = sTRUNC;
    //assert (dt>0); // hopefully.
    assert (dt < _sim->_dt0); // hopefully.
    if(dt<OPT::dtmin) { // time step to small or even 0 
      ::error(bWARNING,"sock step not accepted at %g, dt needed %g is smaller than DTMIN\n", reftime, dt);
      ret=sTROUBLE;
    } else if(dt>0) {
      dt *= stepno; // stepno is valid for next step.
    } else if (dt==0) {
      dt=OPT::dtmin;
      ::error(bWARNING,"dt is 0 set to dtmin %g \n", dt);
      ret=sTROUBLE;
    } else { untested();
        ::error(bDANGER,"dt  in else path should not happen dt %g \n", dt);
    }
    _sim->_time0 = reftime;
    tr_reject();
  }

  if (has_sol_nan("leaving transtep ")) return sTROUBLE;

  //#ifndef NDEBUG
//   for (unsigned i=1; i <= n_vars; i++) {
//     if(std::isnan(_sim->_i[i])||std::isnan(_sim->_vdcstack.top()[i]) ) {
// #ifndef NDEBUG
//       _error = 1;
//       assert(!tr_converged);
// #endif
//       ::error(bDANGER+_verbose,"s_sock: transtep nan in result %d th current %g / variable %g aborting transtep\n",i,_sim->_i[i],_sim->_vdcstack.top()[i]);
//       ret=sTROUBLE;
//       dt=dt/4.0;
//     }
//   }

  _dthack = dt;
  return ret; // FIXME: proper status.
}
/*--------------------------------------------------------------------------*/
void SOCK::tr_reject()
{
  ::status.accept.start();
  _sim->_acceptq.clear();
  ++tr_steps_rejected_;
  ::status.accept.stop();
}
/*--------------------------------------------------------------------------*/
void SOCK::verakons_reply()
{
  stream << ((uint16_t)_error); stream.pad(6);
  for (unsigned i=1; i <= n_vars; i++) {
    trace1("SOCK::kons_reply", _sim->vdc()[i]);
    stream << _sim->vdc()[i];
  }

  //
  // now i contains the negative sum of the cap value.
  // does this break anything?

  for (unsigned i=1; i <= n_vars; i++)         /* Q-Punkt == I == RS */
  {
    trace1("SOCK::verakons_reply  dqdt", _sim->_i[i]);
    stream << _sim->_i[i];
  }

  if(_dump_matrix){
    _out << "u,dq=i: \n";
    for (unsigned i=0; i <= n_vars; i++)         /* Variablen-Werte */
    {
      _out << _sim->vdc()[i] << "," << _sim->_i[i] <<  "\n" ;
    }
  }

  trace0("SOCK::verakons_reply  ac_snap");
  ac_snapshot(); // FIXME: to verakons()

  send_matrix();

  assert(stream.bufsize() >= total);
  stream << SocketStream::eol;

//  for( unsigned i = 0; i < _caplist.size(); i++)
//  {
//    trace1("unlatching", i);
//
//    dashier umgekehrt.
//    commonstash[i] = _caplist[i]->common();
//    detach(_caplist[i]);
//
//    COMMON_COMPONENT* c = bm_dispatcher.clone("eval_bm_value");
//    COMMON_COMPONENT* dc = c->deflate();
//
//    _caplist[i]->attach_common(dc);
//
//    _caplist[i]->keep_ic(); // latch voltage applied to _v0
//
//
//  }

}

void SOCK::transtep_reply(unsigned ret, bool eol)
{
  _error = 0;
  stream << _error; stream.pad(6);
  stream << (uint64_t)ret;
  stream << _dthack;

  for (unsigned i=1; i <= n_vars; i++) {
    trace2("SOCK::transtep_reply  v ", _sim->_vdcstack.top()[i],i);
    stream << _sim->_vdcstack.top()[i];
  }

  // OUCH, move to main loop.
  if(eol) {
    stream << SocketStream::eol;
  }
}

void SOCK::transtep_gc_reply(unsigned ret)
{ itested();
  transtep_reply(ret, false);

  for (unsigned i=1; i <= n_vars; i++)
  {
    trace1("SOCK::transtep_gc_reply  dqdt", _sim->_i[i]);
    stream << _sim->_i[i];
  }

  trace0("SOCK::transtep_gc_reply  ac_snap");
  ac_snapshot(); // BUG: calls tr_begin
  send_matrix();
  assert(stream.bufsize() >= total);
  stream << SocketStream::eol;
}
/*--------------------------------------------------------------------------*/
void SOCK::verainit_reply()
{
  trace4("SOCK::verainit_reply ", _error, n_vars, length, var_namen_total_size);
  stream << _error;   stream.pad(6);
  stream << int32_t (n_vars);  stream.pad(4);
  stream << int32_t (var_namen_total_size);  stream.pad(4);

  assert(stream.tcur() == 24);

  for (unsigned i = 0; i < n_vars; i++) {
    trace2("SOCK::verainit_reply",  i, var_namen_arr[i]);
    putstring8( &stream, var_namen_arr[i]);
    stream << '\t'; stream.pad(7);
  }

//good idea?? not yet. vera insists on branch node
//
//  for (unsigned i=0; i < n_inputs; i++)
//  {
//    trace1("putting name " +  input_names[i], i);
//    putstring8( &stream, input_names[i]);
//    stream << '\t'; stream.pad(7);
//  }
  stream.flush();

  trace0("done verainit_reply");
}
/*--------------------------------------------------------------------------*/
void SOCK::veraop_reply()
{
  assert(n_vars==_sim->_total_nodes);
  trace1("SOCK::veraop_reply ", n_vars);
  stream << _error; stream.pad(6);
  for (unsigned i=1; i <= n_vars; i++)         /* Variablen-Werte */
  {
    stream << _sim->vdc()[i];
  }
  if(_dump_matrix){
    _out << "i,u: \n";
    for (unsigned i=0; i <= n_vars; i++)         /* Variablen-Werte */
    {
      _out << _sim->_i[i] << "," <<  _sim->vdc()[i] << "\n" ;
    }
  }

  trace0("SOCK::veraop_reply taking ac snapshot (matrix only?).");
  ac_snapshot();
  send_matrix();
  stream << SocketStream::eol;

  trace0("SOCK::veraop_reply sent");

  total = 2*n_vars_square+n_vars+1;
  assert(stream.bufsize() >= total);
  if (printlevel >= 1)
  {
    userinfo(1,"vera_titan_ak","Sende: Error %i Framenumber %i, Laenge %i\n",
        _error,frame_number,total);
  }
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void SOCK::send_matrix()
{
  const BSMATRIX<double> G = _sim->_acx.real();
  const BSMATRIX<double> C = _sim->_acx.imag();
  if(_dump_matrix){
     _out << "G\n" << G << "\n";
     _out << "C\n" << C << "\n";
  }
  trace0("SOCK::send_matrix G");
  for (unsigned i=1; i <= n_vars; i++){
    for (unsigned j=1; j <= n_vars; j++) {
      stream << G.s(j,i);
    }
  }
  trace0("SOCK::send_matrix C");
  for (unsigned i=1; i <= n_vars; i++){
    for (unsigned j=1; j <= n_vars; j++) {
      stream << C.s(j,i);
    }
  }
}
/*--------------------------------------------------------------------------*/
void SOCK::cap_prepare(void)
{
  trace1("SOCK::cap_prepare", _caplist.size() );
  assert(!_capstash);
  _capstash = new CARDSTASH[_caplist.size()];

  for (unsigned ii = 0;  ii < _caplist.size();  ++ii) {
    _caplist[ii]->inc_probes();			// we need to keep track of it
    _capstash[ii] = _caplist[ii];			// stash the std value

    if(_caplist[ii]->has_common()){
      trace1("SOCK::cap_prepare. have common", _caplist[ii]->value());
//      _caplist[ii]->set_value(_caplist[ii]->value(), 0); // zap out extensions
//      _caplist[ii]->set_constant(false);		 // update HACK?
//
//    all devices need individual commons.
      _caplist[ii]->attach_common(_caplist[ii]->common()->clone());
      assert(_caplist[ii]->has_common());
    }else{
      //      _sweepval[ii] = _zap[ii]->set__value();	// point to value to patch
      COMMON_COMPONENT* c = bm_dispatcher.clone("eval_bm_value");
      double capval = _caplist[ii]->value();
      trace2("SOCK::cap_prepare, attaching common", _caplist[ii]->long_label(), capval);
      c->set_value(capval);
      COMMON_COMPONENT* dc = c->deflate();
      assert(dc);
      //
      _caplist[ii]->set_value(_caplist[ii]->value(), dc);	// zap out extensions
      _caplist[ii]->set_constant(false);		// so it will be updated
      trace1("SOCK::cap_prepare calling precalc_first", _caplist[ii]->long_label());
      _caplist[ii]->precalc_first();
      trace0("SOCK::cap_prepare calling precalc_last");
      _caplist[ii]->precalc_last();
      trace0("SOCK::cap_prepare calling trbegin");
      _caplist[ii]->tr_begin();
    }
  }
}
/*--------------------------------------------------------------------------*/
void SOCK::cap_reset(void)
{
  trace0("SOCK::cap_reset");
  for (unsigned ii = 0;  ii < _caplist.size();  ++ii) {
      _capstash[ii].restore();
      _caplist[ii]->dec_probes();
//      _caplist[ii]->precalc_first();
//      _caplist[ii]->precalc_last();
  }
  delete[] _capstash;
  _capstash = NULL;
  trace0("SOCK::cap_reset done");
}
/*--------------------------------------------------------------------------*/



bool SOCK::relax_kons(bool converged_in)
{
  bool converged=converged_in;
  OPT::ITL itl=OPT::DCBIAS;
  // === BEGIN a mini step for relaxing the strong voltages: ===
  _sim->_uic = false;
  trace4("SOCK::relax_kons ",_sim->_dt0,_sim->_time0, _sim->_phase, _sim->_mode);
  _sim->_dt0 = OPT::dtmin;
  _sim->_time0 += _sim->_dt0;
  trace4("SOCK::relax_kons ",_sim->_dt0,_sim->_time0, _sim->_phase, _sim->_mode);
  advance_time(); // sink _time0 into devices
  trace4("SOCK::relax_kons ",_sim->_dt0,_sim->_time0, _sim->_phase, _sim->_mode);
  _sim->set_command_tran();
  if (printlist().size()) {
    outdata(_sim->_time0, ofPRINT);
  }
  _sim->_phase = p_TRAN;
  bool tr_converged;
  try {
    // for( unsigned i = 0; i < _caplist.size(); i++) { itested();
    //   //assert(!_caplist[i]->is_constant());
    //   _caplist[i]->q_eval(); 
    // }
    trace1("SOCK:relax_kons solving for mini time step",_sim->_dt0);
    tr_converged = solve(OPT::TRHIGH, _trace);
  }catch (Exception e) { incomplete();
    ::error(bDANGER+_verbose, "solve for dtmin time step failed in relax_kons\n");
    throw e;
  }
  if(_dump_matrix){
    _out << "A u,dq=i: \n";
    for (unsigned i=0; i <= n_vars; i++)         /* Variablen-Werte */
    {
      _out << _sim->_v0[i] << "," << _sim->_i[i] <<  "\n" ;
    }
  }

  ::status.accept.start();
  CARD_LIST::card_list.tr_accept();
  ::status.accept.stop();
  if(_dump_matrix){
    _out << "B u,dq=i: \n";
    for (unsigned i=0; i <= n_vars; i++)         /* Variablen-Werte */
    {
      _out << _sim->_v0[i] << "," << _sim->_i[i] <<  "\n" ;
    }
  }
  // ====== END a mini step for relaxing the strong voltages =====

  // ===== do cons again 


  //        n_eingaenge == #caps?
  _sim->set_command_dc();
  trace1("SOCK::kons", _caplist.size());
  _sim->_time0 = 0; // just assert?
  _sim->_dt0 = 0; // just assert?
  trace0("SOCK::kons advance_time");
  advance_time();
  trace1("SOCK::kons ",_sim->_phase);
  _sim->_phase = p_INIT_DC;
  total =  n_eingaenge + n_vars + 1;


  assert(_sim->_v0);
  for (unsigned i=1; i <= n_vars; i++) {
    //_sim->_v0[i] =_sim->vdc()[i]
    trace2("SOCK::kons start ", i,  _sim->_v0[i] );
  }
  if (printlist().size()) {
    outdata(0., ofPRINT);
  }
  _sim->keep_voltages(); // v0->vdc

  _error = 0; /* verakons(Dwork, x_new, q_dot, G, C); */
  //	n_vars = A->n_var;

  for( unsigned i = 0; i < _caplist.size(); i++) {
    //trace1("SOCK::kons",_caplist[i]->long_label());
    ::error(bDEBUG+_verbose,
            "SOCK::kons keep ic for %s \n",_caplist[i]->long_label().c_str());
    _caplist[i]->keep_ic(); // latch voltage applied to _v0
    // this sets has_ic() to true which is a problem
    // for using ther ic given in the netlist
    // and for the next operating point. 
    _caplist[i]->set_constant(true); // maybe not a good idea at all.
    _caplist[i]->q_eval();		// so it will be updated
  }
  //
  //================ 2. dc========
  _sim->_uic = true;
  _sim->_bypass_ok = false;
  _sim->set_inc_mode_bad();

  itl = OPT::DCBIAS;

  if(printlist().size()) {
    head(0,0," ");
  }
  CARD_LIST::card_list.tr_begin();
  converged = false;
  try{
    for( unsigned i = 0; i < _caplist.size(); i++) {
      //trace1("SOCK::kons",_caplist[i]->long_label());
      if(_verbose > 10) {
        ::error(bDEBUG+_verbose,"SOCK::kons: setting cap %s \n",
                (_caplist[i]->long_label()).c_str());
      }
      // assert(!_caplist[i]->is_constant()); // NO caps are const, when they are in VS mode.
      _caplist[i]->do_tr(); // so it will be updated. is this sufficient?
			     // is it sufficient to only queue caps?
    }
    trace1("SOCK::verakons: solving",_trace);
    converged = solve(itl,_trace);
    //    solve(OPT::TRHIGH,_trace);
    //    solve_with_homotopy(itl,_trace);
    // homotopy is to much effort calling
    // procedures must catch problems with convergence.
  }catch(Exception e) { untested();
    ::error(bDANGER+_verbose, "solve failed in verakons\n");
    throw e;
  }
  if(!converged){
    ::error(bWARNING, "s_sock::verakons: solve did not converge\n");
    converged = solve_with_homotopy(itl,_trace);
  }
  if(!converged){
    _error = 1;
  }
  ::status.accept.start();
  assert(_sim->_uic);
  CARD_LIST::card_list.tr_accept();
  ::status.accept.stop();
  
  if(_dump_matrix){
    _out << "AA u,dq=i: \n";
    for (unsigned i=0; i <= n_vars; i++)         /* Variablen-Werte */
    {
      _out << _sim->_v0[i] << "," << _sim->_i[i] <<  "\n" ;
    }
  }

  // == end 2. dc 
  return converged;
}
/*--------------------------------------------------------------------------*/

bool SOCK::has_sol_nan(string message)
{
  //#ifndef NDEBUG
  for (unsigned i=1; i <= n_vars; i++) {
    if(std::isnan(_sim->_i[i])||std::isnan(_sim->_vdcstack.top()[i]) ) {
#ifndef NDEBUG
      _error = 1;
#endif
      ::error(bDANGER+_verbose,"WARNING: s_sock: %s nan in result %d th current %g / variable %g \n",message.c_str(),i,_sim->_i[i],_sim->_vdcstack.top()[i]);
      return true;
    }
    if(fabs(_sim->_i[i]) > 1E50) {
      ::error(bDANGER+_verbose,"WARNING: s_sock: %s %d th. right hand side of matrix has too large value %g \n",message.c_str(),i,_sim->_i[i]);
      return true;
    }
    if(fabs(_sim->_vdcstack.top()[i]) > 1E50) {
      ::error(bDANGER+_verbose,"WARNING: s_sock: %s %d th. variable has too large value  %g \n",message.c_str(),i,_sim->_i[i]);
      return true;
    }
    
  }
  return false;
}




SOCK::~SOCK()
{
  trace0("SOCK::~SOCK()");
  if(_status){
    delete(_status);
  }else{
  }
}
/*--------------------------------------------------------------------------*/
// vim:ts=8:sw=2:noet:

